package as.adamsmith.etherealdialpad.dsp;

import android.app.*;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.IBinder;
import android.os.Process;
import android.os.RemoteException;
import android.preference.PreferenceManager;
import android.util.Log;

public class SynthService extends Service {
	private static final String TAG = "SynthService";
	
	Thread thread;
	
	@Override
	public IBinder onBind(Intent intent) {
		final SharedPreferences synthPrefs = PreferenceManager
			.getDefaultSharedPreferences(getApplicationContext());
		
		final float[] scale = buildScale(synthPrefs.getString("quantizer", "1,4,6,9,11"));
		final int octaves = Integer.parseInt(synthPrefs.getString("octaves", "3"));
		
		final boolean delay = synthPrefs.getBoolean("delay", true);
		final boolean softTimbre = synthPrefs.getBoolean("softTimbre", true);
		final boolean softEnvelope = synthPrefs.getBoolean("softEnvelope", true);
		final boolean duet = synthPrefs.getBoolean("duet", true);
    	
		final WtOsc ugOscA1 = new WtOsc();
		final WtOsc ugOscA2 = new WtOsc();
		final ExpEnv ugEnvA = new ExpEnv();
	    
		final WtOsc ugOscB1 = new WtOsc();
		final WtOsc ugOscB2 = new WtOsc();
		final ExpEnv ugEnvB = new ExpEnv();
	    
		if(softTimbre) {
			ugOscA1.fillWithHardSin(7.0f);
			ugOscB1.fillWithHardSin(7.0f);
			ugOscA2.fillWithHardSin(2.0f);
			ugOscB2.fillWithHardSin(2.0f);
		} else {
			ugOscA1.fillWithSqrDuty(0.6f);
			ugOscB1.fillWithSqrDuty(0.6f);
			ugOscA2.fillWithSqrDuty(0.9f);
			ugOscB2.fillWithSqrDuty(0.9f);
		}
		
		final Dac ugDac = new Dac();
	    
	    if(delay) {
	    	final Delay ugDelay = new Delay(UGen.SAMPLE_RATE/2);
	    	ugEnvA.chuck(ugDelay);
	    	ugEnvB.chuck(ugDelay);
	    	ugDelay.chuck(ugDac);
	    } else {
	    	ugEnvA.chuck(ugDac);
	    	ugEnvB.chuck(ugDac);
	    }
	    
	    if(!softEnvelope) {
	    	ugEnvA.setFactor(ExpEnv.hardFactor);
	    	ugEnvB.setFactor(ExpEnv.hardFactor);
	    }
		
		if(thread != null) {
			Log.e(TAG,
					"Someone tried to bind when there was already an audio thread!",
					new RuntimeException());
		}
		
		thread = new Thread() {
			@Override
			public void run() {
				Log.d(TAG, "started audio rendering");
			    
				System.gc();
				Process.setThreadPriority(Process.THREAD_PRIORITY_AUDIO);
		        
				ugDac.open();
		        
				while(!isInterrupted()) {
					// TODO: apply control!
					
					ugDac.tick();
				}
				
				ugDac.close();
				
				Log.d(TAG, "finished audio rendering");
			}
		};
		
		thread.start();
		
		return new ISynthService.Stub() {

			@Override public void up() throws RemoteException {}
			@Override public void down() throws RemoteException {}
			@Override public void left() throws RemoteException {}
			@Override public void right() throws RemoteException {}
			
			@Override public void select() throws RemoteException {}

			@Override public void metaOn() throws RemoteException {}
			@Override public void metaOff() throws RemoteException {}
			
			@Override public void primaryOff() throws RemoteException {
				ugEnvA.setActive(false);
			}

			@Override public void primaryOn() throws RemoteException {
				// lazily connect
				ugOscA1.chuck(ugEnvA);
				if(duet) {
					ugOscA2.chuck(ugEnvA);
				}
				
				ugEnvA.setActive(true);
			}

			@Override public void primaryXY(float x, float y) throws RemoteException {
				ugOscA1.setFreq(buildFrequency(scale, octaves, x));
				if(duet) {
					ugOscA2.setFreq(buildFrequency(scale, octaves, y));
				} else {
					ugEnvA.setGain(2*y*y);
				}
			}

			@Override public void secondaryOff() throws RemoteException {
				ugEnvB.setActive(false);
			}

			@Override public void secondaryOn() throws RemoteException {
				// lazily connect
				ugOscB1.chuck(ugEnvB);
				if(duet) {
					ugOscB2.chuck(ugEnvB);
				}
				
				ugEnvB.setActive(true);
			}

			@Override public void secondaryXY(float x, float y) throws RemoteException {
				ugOscB1.setFreq(buildFrequency(scale, octaves, x));
				if(duet) {
					ugOscB2.setFreq(buildFrequency(scale, octaves, y));
				} else {
					ugEnvB.setGain(2*y*y);
				}
			}
		}; 
	}
	
	
	@Override
	public void onDestroy() {
		thread.interrupt();
		try {
			thread.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	
	static float[] buildScale(String quantizerString) {
		if(quantizerString != null && quantizerString.length() > 0) {
			String[] parts = quantizerString.split(",");
			float[] scale = new float[parts.length];
			for(int i = 0; i < parts.length; i++) {
				scale[i] = Float.parseFloat(parts[i]);
			}
			return scale;
		} else {
			return null;
		}
	}
	
	static float buildFrequency(float[] scale, int octaves, float input) {
		final float base = 48;
		
		if(scale == null) {
			float mapped = base + input*octaves*12.0f;
			return (float)Math.pow(2, (mapped-69.0f)/12.0f) * 440.0f;
		} else {
			int idx = (int)(scale.length*input*octaves);
			float mapped = base + scale[idx%scale.length] + 12*(idx/scale.length);
			return (float)Math.pow(2, (mapped-69.0f)/12.0f) * 440.0f;
		}
		
	}
}
